
// ahrs.c
#include <stdio.h>
#include <math.h>
#include "ahrs.h"
#include "sensor_cal.h"
#include "hmc5843.h"
#include <errno.h> // for errno

/*
Estimator / Extended Kalman filter IMU application

Filter concept from Gebre-Egziabher paper "Design of Multi-Sensor Attitude Determination Systems"
Estimator concept from Elkaim paper "MAV Attitude Determination by Vector Matching"

Input Requirements
Gyro measurements gyro_roll_rps(), gyro_pitch_rps(), gyro_yaw_rps()
Accelerometer measurements accelero_x(), accelero_y(), accelero_z()
Magnetometer measurements magx(), magy(), magz()

matrix format is
[ 0,0  0,1  0,2  0,3]
[ 1,0  1,1  1,2	 1,3]
[ 2,0  2,1  2,2  2,3]
[ 3,0  3,1  3,2  3,3]
*/

# define tau 1000.		// gyro time constant in seconds
# define m1otau -1./tau
# define alpha_estimator .3 // estimator convergence factor (something .1 to 1)

//Covariance matrix for state update
m_t P= {												
	{ 0.01,	0.,		0.,		0.,			0.,			0.		},	// initial confidence
	{ 0.,	0.01,	0.,		0.,			0.,			0.		},	// low number is Hi confidence
	{ 0.,	0.,		0.01,	0.,			0.,			0.		},	
	{ 0.,	0.,		0.,		0.0001,		0.,			0.		},	
	{ 0.,	0.,		0.,		0.,			0.0001,		0.		},	
	{ 0.,	0.,		0.,		0.,			0.,			0.0001	}	
	};

// estimator (aiding system) quaternion is estimated attitude from accelerometers, magnetometers, gps
v_t qhat = {1.,0.,0.,0.};

// state quaternion attitude
v_t qstate = {1.,0.,0.,0.};

// gyro bias vector
v_t bias = {0.,0.,0.};


// ahrs init - assume we have some good sensor values in buffers
// initialize estimator and state quaternions 
// initialize gyro bias from the initial gyro measurements.
void ahrs_init()	
{
	int i; 				// for loop

// do some estimator updates to converge estimator
	for (i=0; i < 20; i++) estimator_update();

// initialize qstate with estimator quaternion conjugate
	qstate[0] = qhat[0];
	qstate[1] = -qhat[1];
	qstate[2] = -qhat[2];
	qstate[3] = -qhat[3];

// initialize gyro biases with current (assumed static) gyro reading
	bias[0]= gyro_roll_rps();
	bias[1]= gyro_pitch_rps();
	bias[2]= gyro_yaw_rps();

} //end ahrs_init


// Kalman time update
// propegate qstate from gyro measurements and dt using qstate = q + dt( omega(w-b) )q
// propegate covariance P = FPF' +Cd							
void time_update(f_t dt)
{
	m_t			Ft;		// 6x6 F'
	m_t			PFt;	// 6x6 PF'
	v_t			qtmp;	// 4
	
// Cd derived in matlab from GQwG'
// G is a function of alpha and beta (both set to 1)
// Qw top left 3 diagonals are (sigma p,q,r)^2.  sigma p,q,r = .05deg/sec = .0009 rad/sec
// Qw bot right 3 diagonals are 2 * sigmaw^2/tauw.  sigmaw = = .05deg/sec = .0009 rad/sec.  tauw = 1000sec
// expressed here as a vector of diagonals
v_t Cd = { .81e-06, .81e-06, .81e-06, .0016e-06, .0016e-06, .0016e-06 };		
	
	f_t		p = gyro_roll_rps(); // measured gyro rate to simplify calculations
	f_t		q = gyro_pitch_rps();
	f_t		r = gyro_yaw_rps();

   	f_t		wmb1 = (p - bias[0])/2.; // (measured gyro rate - bias)/2 to simplify calculations
	f_t		wmb2 = (q - bias[1])/2.;
	f_t		wmb3 = (r - bias[2])/2.;

// Construct Omega/2 (4x4) from the 3 gyro measurements and the bias estimates

m_t Omega = {	{	0.,		-wmb1,	-wmb2, 	-wmb3 	},
				{	wmb1,	 0.,	 wmb3, 	-wmb2 	},
				{	wmb2,	-wmb3,	 0., 	 wmb1 	},
				{	wmb3,  	 wmb2,	-wmb1, 	 0. 	}};

	// propegate qstate 
	mv_mult(qtmp,Omega,qstate,4,4);		// qtmp =  Omega(w-b)qstate
	v_scale (qtmp,qtmp,dt,4);		// times dt
	v_add(qstate,qstate,qtmp,4);	// qstate = qstate + qtmp = qstate + Omega(w-b)qstate
	v_normq( qstate, qstate );		// normalize

// Build the F matrix to propegate covariance
// top left 3x3 is [-wcross]	top right is .5*[I3x3]
// lower left is [I3x3]			lower right is -1/tau * [I3x3]
m_t F = {	{	0.,		-r,		 q,		.5,			0.,			0.		},
			{	 r,		0.,		-p,		0.,			.5,			0.		},
			{	-q,		 p,		0.,		0.,			0.,			.5		},
 			{	1.,		0.,		0.,		m1otau,		0.,			0.		},
			{	0.,		1.,		0.,		0.,			m1otau,		0.		},
			{	0.,		0.,		1.,		0.,			0.,			m1otau	}};

	// do the calcs
	
	m_scale(F,F,dt,6,6);				// F = F * dt (whole matrix multiplied by dt!!!)
	m_transpose( Ft, F, 6, 6 );  		// Ft = F'
	m_mult( PFt, P, Ft, 6, 6, 6 ); 		// PFt = PF'
	m_mult( P, F, PFt, 6, 6, 6 ); 		// P = FPF'
	mv_add_diag( P, Cd, 6 );			// P = P + Cd = FPF' +Cd
	
	//m_add( P, P, Cd, 6, 6 ); 			// P = P + Cd = FPF' +Cd

} // end time_update


// Kalman measurement update 
// update qstate and bias vector
// update covariance with P = P - LHP
void measurement_update()
{	m_t Hkt;	// 6x3 H'
	m_t HkP;	// 3x6 Hk P
	m_t PHkt;	// 6x3 P Hk'
	m_t HkPHkt;	// 3x3 Hk P Hk'
	m_t	S;		// 3x3 inv(Hk P Hk'+R)
	m_t	L;		// 6x3 Kalman gain
	v_t	qe;		// error quaternion
	v_t dx;		//
	v_t	dy;		//

// R is wideband noise on the measurements of p, q, and r from the estimator.
// sigmaphi = .25 deg/sec =  .0044 rad/sec.  Diagonals are square of these
// here sigmaphi = sigmatheta = sigmapsi
// expressed here as a vector of diagonals
v_t R = {.00002, .00002, .00002  };			

// build measurement matrix Hk [I3x3 zero3x3] and transpose Hk'

m_t Hk = {	{1.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.}};

// calculate kalman gain L = PH' * inv(HPH'+R)
	m_transpose( Hkt, Hk, 3, 6 );				// Ht[6,3] = (transpose)H[3,6]	
	m_mult( PHkt, P, Hkt, 6, 6, 3 );		// PHkt[6,3] = P[6,6] * Hkt[6,3]			P Hk'
	m_mult( HkPHkt, Hk, PHkt, 3, 6, 3 ); 	// HkPHkt[3,3] = Hk[3,6] * PHkt[6,3] 		Hk P Hk'
	mv_add_diag( HkPHkt, R, 3 );			// HkPHkt[3,3] = HkPHkt[3,3] + R[3,3] 	Hk P Hk '+ R
	m33_inv( S, HkPHkt );					// S[3,3] = INV(HkPHkt[3,3])				inv(Hk P Hk'+R)
	m_mult( L, PHkt, S, 6, 3, 3 );			// L[6,3] = PHkt[6,3] * S[3,3] 			PHk' * inv(Hk P Hk'+R)
	
	// calculate error quaternion difference between present state and estimator value
	quat_mult( qe, qhat, qstate );

	dy[0] = -qe[1];	// dy is vector portion of error
	dy[1] = -qe[2];
	dy[2] = -qe[3];

	// calculate dx vector portion
	mv_mult( dx, L, dy, 6, 3 );	// dx[6] = L[6,3] * dy[3] note dx is [6] not [7]

	// update gyro bias
	bias[0]-= dx[3];
	bias[1]-= dx[4];
	bias[2]-= dx[5];

	// express as full quaternion
	dx[3]=dx[2];
	dx[2]=dx[1];
	dx[1]=dx[0];
	dx[0]=1.;

	quat_mult( qstate, qstate, dx);	// update state quaternion
	v_normq( qstate, qstate );		// normalize

	// update covariance P = P - LHP
	
	m_mult( HkP, Hk, P, 3, 6, 6 );	// HkP[3,6] = Hk[3,6] * P[6,6]		Hk P
	m_mult( S, L, HkP, 6, 3, 6 );	// S[6,6] = L[6,3] HkP[3,6]			L Hk P
	m_sub( P, P, S, 6, 6 );			// P[6,6] = P[6,6] - S[6,6]		P = P - S = P - L Hk P

} // end measurement_update


void estimator_update() // least squares implementation of vector matching
{		
		//iter_t i;		// for loops
		v_t	mb; 		// mag vector measured in body frame
		v_t mnhat;		// calculated in nav frame
		v_t	ab; 		// measured gravity vector in body frame
		v_t anhat;		// calculated in nav frame

		v_t delta;		// vector of errors
		//m_t H;  		// observation matrix
		m_t Ht;			// transpose H
		m_t HtH; 		// Ht*H
		m_t HtHinv;		// HtH inverse
		m_t Hpinv;		// pseudo inverse of H
		v_t qe;			// estimated error in nav frame

		//v_t mn = { 24027.54, -2504.11, 40939.86 };	// known local mag field in nav frame from noaa data
		//v_t an = ( 0., 0., 9.8 };						// known local gravity in nav frame
		//v_norm(mn,mn); // normalize
		//v_norm(an,an);

		v_t mn = { 0.50546, -0.05268, 0.86124 };	// normalized known local mag field in nav frame
		v_t an = { 0., 0., 1. };					// normalized known local gravity in nav frame

		mb[0]= magx();			// measured mag field in body frame
		mb[1]= magy();
		mb[2]= -magz();
		v_norm(mb,mb);			// normalize measurement

		ab[0]= accelero_x();	// measured acceleration in body frame
		ab[1]= -accelero_y();
		ab[2]= -accelero_z();
		v_norm(ab,ab);			// normalize measurement

	//for (i=0; i<50; i++) // if multiple iterations are used
	{

		// rotate measurements by attitude estimate (body frame to nav frame)
		v_rotate(mnhat,mb,qhat); 	
		v_rotate(anhat,ab,qhat);

		// these are in the nav frame - here we can correct anhat with gps and baro altitude accelerations

		// calculate errors (in nav frame)
		delta[0] = an[0] - anhat[0]; // known vector - measurement rotated
		delta[1] = an[1] - anhat[1]; 
		delta[2] = an[2] - anhat[2];
		delta[3] = mn[0] - mnhat[0];
		delta[4] = mn[1] - mnhat[1];
		delta[5] = mn[2] - mnhat[2];

	// build observation matrix H
	// we'll use 2*anhat and 2*mnhat in the matrix so multiply now
		v_scale(anhat,anhat,2.,3);
		v_scale(mnhat,mnhat,2.,3);

	// accelerometer part is top 3x3	anhat cross
	// magnetometer part is bottom 3x3	mnhat cross

m_t H = {	{	       0.,	-anhat[2],	 anhat[1]	},
			{	 anhat[2],	       0.,	-anhat[0]	},
 			{	-anhat[1],	 anhat[0],	       0.	},
 			{	       0.,	-mnhat[2],	 mnhat[1]	},
			{	 mnhat[2],	       0.,	-mnhat[0]	},
 			{	-mnhat[1],	 mnhat[0],	       0.	}};
	

		m_transpose( Ht, H, 3, 6 );		// Ht[3,6] = (transpose)H[6,3]
		m_mult(HtH,Ht,H,3,6,3);			// HtH[3,3] = Ht[3,6] * H[6,3]			Ht will be used later also
		m33_inv( HtHinv, HtH );			// HtHinv is now inv(HtH)[3,3]
		m_mult(Hpinv,HtHinv,Ht,3,3,6);	// Hpinv[3,6] is now the pseudo inverse
		m_scale(H, Hpinv, alpha_estimator, 6, 3);	// scale by alpha = .3 (something from .1 to 1)
		mv_mult( qe, H, delta, 3, 6 );	// qe[3] = H[3,6] * deltaan[6] 

		qe[3]=qe[2];	// relocate vector positions to form quaternion
		qe[2]=qe[1];
		qe[1]=qe[0];
		qe[0]=1.;

		quat_mult(qhat,qhat,qe); 	// qhat = qhat x qe (quaternion multiplication)
		v_normq(qhat,qhat);			// adds the calculated error in rotation to the qhat estimate

	} 	// next i
}		// end estimator_update


// a few routines to retrieve local data

// extracts attitude in angles from quaternion qstate
void getangles(v_t angles_out)
{	quat2euler(angles_out,qstate);	} 

// extracts attitude in angles from estimator quaternion qhat
void getestimator(v_t angles_out)
{	v_t	qas; // tmp conjugate used for display
	qas[0] = qhat[0];
	qas[1] = -qhat[1];
	qas[2] = -qhat[2];
	qas[3] = -qhat[3];
	quat2euler(angles_out,qas);	} 

// extract bias
void getbias(v_t bias_out)
{
bias_out[0] = bias[0];
bias_out[1] = bias[1];
bias_out[2] = bias[2];
}

// get state quaternion
void getqstate(v_t q_out)
{	q_out[0]=qstate[0];	
	q_out[1]=qstate[1];	
	q_out[2]=qstate[2];	
	q_out[3]=qstate[3];	
} 

//get estimator quaternion
void getqhat(v_t q_out)
{	q_out[0]=qhat[0];	
	q_out[1]=qhat[1];	
	q_out[2]=qhat[2];	
	q_out[3]=qhat[3];	
} 

