// vector, quaternion, matrix library for ahrs
// collected by Mitch


#include <stdio.h>
#include <math.h>
#include <errno.h> // for errno
#include "vqm_lib.h"

// This will normalize a quaternion vector q
// q/norm(q)    q(4,1)
void v_normq(
	v_t	q_out,
const	v_t	q
)
{
	v_scale(
		q_out,
		q,
		1.0 / sqrt( q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3] ),
		4
	);
}


void quat_mult(v_t qout, const v_t q1, const v_t q2)
//Given two quaternions q1 and q2, calculate the combined rotation qout
{
    qout[0] = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3];
    qout[1] = q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2];
    qout[2] = q1[0] * q2[2] - q1[1] * q2[3] + q1[2] * q2[0] + q1[3] * q2[1];
    qout[3] = q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1] + q1[3] * q2[0];
    
    //v_normq(qout, qout);
}

// v_rotate
// rotate the vector(3) by the unit quaternion(4) using vout = qin' * vin * qin
// Given a quaternion qin = qin[0] + qin[1]i + qin[2]j + qin[3]k (with |z| = 1) 
// and a vector vin with coordinates vin[0], vin[1], and vin[2], 
// the following code performs a rotation to produce the vector vout

void v_rotate(v_t vout, const v_t vin, const v_t qin)
{ 	/*
	f_t	t2 =   qin[0] * qin[1];
 	f_t	t3 =   qin[0] * qin[2];
 	f_t	t4 =   qin[0] * qin[3];
 	f_t	t5 =  -qin[1] * qin[1];
 	f_t	t6 =   qin[1] * qin[2];
 	f_t	t7 =   qin[1] * qin[3];
 	f_t	t8 =  -qin[2] * qin[2];
 	f_t	t9 =   qin[2] * qin[3];
 	f_t	t10 = -qin[3] * qin[3];
	*/
	
	// opposite direction rotation qin * vin * qin'
	// take negatives of qin 1,2,3
	f_t	t2 =  -qin[0] * qin[1];
 	f_t	t3 =  -qin[0] * qin[2];
 	f_t	t4 =  -qin[0] * qin[3];
 	f_t	t5 =  -qin[1] * qin[1];
 	f_t	t6 =   qin[1] * qin[2];
 	f_t	t7 =   qin[1] * qin[3];
 	f_t	t8 =  -qin[2] * qin[2];
 	f_t	t9 =   qin[2] * qin[3];
 	f_t	t10 = -qin[3] * qin[3];


 vout[0] = 2*( (t8 + t10)* vin[0] + (t6 -  t4)* vin[1] + (t3 + t7)* vin[2] ) + vin[0];
 vout[1] = 2*( (t4 +  t6)* vin[0] + (t5 + t10)* vin[1] + (t9 - t2)* vin[2] ) + vin[1];
 vout[2] = 2*( (t7 -  t3)* vin[0] + (t2 +  t9)* vin[1] + (t5 + t8)* vin[2] ) + vin[2];
} // end v_rotate


/*
 * This will convert from quaternions to euler angles using Tait Bryan (1,2,3) method
 * q(4,1) -> euler[phi;theta;psi] (rad)
 */
void quat2euler(
	v_t	euler,
	const v_t q
)
{
		errno = 0; //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	f_t			test = q[1]*q[3] - q[0]*q[2];


	if (test > .499){						// singularity at north pole
		euler[0] = 0;						// phi
		euler[1] = -1.5708; 				// theta pi/2
		euler[2] = -2. * atan2(q[1],q[0]);	//psi
		return;
	}	

	if (test < -.499){						// singularity at south pole
		euler[0] = 0;						// phi
		euler[1] = 1.5708;				// theta
		euler[2] = 2 * atan2(q[1],q[0]);	//psi
		return;
	}	
	
	


	f_t			q02 = q[0]*q[0];
	f_t			q12 = q[1]*q[1];
	f_t			q22 = q[2]*q[2];
	f_t			q32 = q[3]*q[3];

	// phi = atan2(DCM[2,3],DCM[3,3])
	euler[0] = atan2( 2.0 * (q[2]*q[3] + q[0]*q[1]), q32 - q22 - q12 + q02); // q32 - q22 - q12 + q02 ); //1. - 2.* (q12 + q22)

	// theta = asin -DCM[1,3]
	euler[1] = -asin( 2.0 * test );

	// psi = atan2(DCM[1,2],DCM[1,1])
    euler[2] = atan2( 2.0 * (q[1]*q[2] + q[0]*q[3]), q12 + q02 - q32 - q22); // q12 + q02 - q32 - q22 ); //1. - 2. * (q22 + q32)

	if (errno) perror("Error quat2euler!!!!!");  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

}
	

// vector operations!!!!!!!!!!!!!!!!!!!!!!!!!!!

// vector zeroize
 void v_zero(v_t c, iter_t n)
{
do {
		--n;
		c[n] = 0.;
	} while( n > 0 );
}

/*
 * This will normalize a vector v
 * v/norm(v)
 * v(3,1)
 */
void v_norm(
	v_t	v_out,
	const v_t	v
)
{
	v_scale(
		v_out,
		v,
		1.0 / sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2]),
		3
	);
}

// vector add
 void v_add(
	v_t 			c,
	const v_t 		a,
	const v_t 		b,
	iter_t			n
)
{
	do {
		--n;
		c[n] = a[n] + b[n];
	} while( n > 0 );
}

// vector subtract
 void v_sub(					//c=a-b
	v_t 			c,
	const v_t 		a,
	const v_t 		b,
	iter_t			n
)
{
	do {
		--n;
		c[n] = a[n] - b[n];
	} while( n > 0 );
}


// vector * scaler
 void v_scale(
	v_t			c,
	const v_t	a,
	const f_t	s,
	iter_t		n
)
{
	do {
		--n;
		c[n] = a[n] * s;
	} while( n > 0 );
}


// Matrix operations !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 
// matrix zeroixe
 void m_zero(m_t c, iter_t m, iter_t n)
{
do {
		--n;
	v_zero( c[n], m );
	} while( n > 0 );
}

/*
 * Adds two matrices: c = a + b
 *
 * Safe to call with c == a and c == b
 * Requires 2300 instructions (580 useconds)
 */
 void m_add(
	m_t 		c,
	m_t 	a,
	m_t	b,
	iter_t		m,
	iter_t		n
)
{
	do {
		n--;
		v_add( c[n], a[n], b[n], m );
	} while( n > 0 );
}

// matrix subtract
 void m_sub(
	m_t 		c,
	m_t 		a,
	m_t			b,
	iter_t		m,
	iter_t		n
)
{
	do {
		n--;
		v_sub( c[n], a[n], b[n], m );		//c=a-b
	} while( n > 0 );
}

//	matrix scale
 void m_scale(
	m_t			c,
	m_t			a,
	const f_t			s,
	iter_t		m,
	iter_t		n
)
{
	do {
		n--;
		v_scale( c[n], a[n], s, m );
	} while( n > 0 );
}

// matrix vector add diagonal
// adds the vector a to corresponding diagonals in matrix b
 void mv_add_diag(
	m_t 			b,
	const v_t 		a,
	iter_t			n
)
{
	do {
		--n;
		b[n][n] = a[n] + b[n][n];
	} while( n > 0 );
}

/*	matrix vector multiply
 *  vc = ma * vb
 * 	Unsafe to call with vc == vb
 */
 void mv_mult(
	v_t			vc,
	m_t			ma,
	const v_t	vb,
	iter_t		m,
	iter_t		n
)
{
	iter_t		i;
	iter_t		j;

	for( j=0 ; j<m ; ++j )
	{
		f_t			sum = (f_t) 0;		
		const f_t *		row = &ma[j][0];

		for( i=0 ; i<n ; ++i )
			sum += row[i] * vb[i];

		vc[j] = sum;
	}
}


// vector matrix multiply
// vc = va * mb
// Unsafe to call with vc == va
 void vm_mult(
	v_t			vc,
	const v_t	va,
	m_t			mb,
	iter_t		m,
	iter_t		n
)
{
	iter_t		i;
	iter_t		j;

	for( i=0 ; i<n ; ++i )
	{
		f_t			sum = (f_t) 0;

		for( j=0 ; j<m ; j++ )
			sum += va[j] * mb[j][i];

		vc[j] = sum;
	}
}

// matrix multiply  c = a * b
// Unsafe for c == a or c == b.
// Optimal form runs in 1/2 the time of the simple form due to
// loop constant hoisting and others.

 void m_mult(
	m_t			c,
	m_t			a,
	m_t			b,
	iter_t		m,
	iter_t		n,
	iter_t		p
)
{
	iter_t		i;
	iter_t		j;
	iter_t		k;
/*
// start simple
	m_zero( c );

	for( i=0 ; i<m ; ++i )
		for( j=0 ; j<p ; ++j )
			for( k=0 ; k<n ; ++k )
				c[i][j] += a[i][k] * b[k][j];
//end simple
*/
//start optimal
	for( i=0 ; i<m ; ++i )
	{
		for( j=0 ; j<p ; ++j )
		{
			f_t		sum = (f_t) 0;
			const f_t *	A = &a[i][0];

			for( k=0 ; k<n ; ++k )
				sum += A[k] * b[k][j];
			c[i][j] = sum;
		}
	}
// end optimal

}


// transpose is safe to call with c == a
 void m_transpose(
	m_t			c,
	m_t			a,
	iter_t		m,
	iter_t		n
)
{
	iter_t		i;
	iter_t		j;

	for( i=0 ; i<m ; ++i )	//i is m
	{
		for( j=i ; j<n ; j++ )	//j is n
		{
			f_t			u = a[i][j];
			f_t			l = a[j][i];

			c[j][i] = u;
			c[i][j] = l;
		}
	}
}

// Produce the determinant of a 3x3 matrix
 f_t m33_det(
	m_t			m
)
{
	return	m[0][0] * m[1][1] * m[2][2]
	- 	m[0][0] * m[1][2] * m[2][1]
	-	m[0][1] * m[1][0] * m[2][2]
	+	m[0][1] * m[1][2] * m[2][0]
	+	m[0][2] * m[1][0] * m[2][1]
	-	m[0][2] * m[1][1] * m[2][0]
	;
}


// m33_inv will invert a 3x3 matrix.
// Unsafe to call with c == a
// Details were copied from:  http://mathworld.wolfram.com/MatrixInverse.html
 void m33_inv(
	m_t		D,
	m_t		m
)
{
	D[0][0] = m[1][1] * m[2][2] - m[1][2] * m[2][1];
	D[0][1] = m[0][2] * m[2][1] - m[0][1] * m[2][2];
	D[0][2] = m[0][1] * m[1][2] - m[0][2] * m[1][1];

	D[1][0] = m[1][2] * m[2][0] - m[1][0] * m[2][2];
	D[1][1] = m[0][0] * m[2][2] - m[0][2] * m[2][0];
	D[1][2] = m[0][2] * m[1][0] - m[0][0] * m[1][2];

	D[2][0] = m[1][0] * m[2][1] - m[1][1] * m[2][0];
	D[2][1] = m[0][1] * m[2][0] - m[0][0] * m[2][1];
	D[2][2] = m[0][0] * m[1][1] - m[0][1] * m[1][0];

	m_scale( D, D, 1.0 / m33_det(m), 3, 3 );
}

// print a matrix
 void m_print(
	const char *		label,
	m_t			a,
	iter_t		m,
	iter_t		n
)
{
	iter_t		i;
	iter_t		j;

	printf( "%s =\n\r", label );
	for( i=0 ; i<m ; i++ )
	{
		printf( "[" );
		for( j=0 ; j<n ; j++ )
			printf( " %lf", a[i][j] );
		printf( "]\n\r" );
	}
	printf( "\n\r" );
}

// print a vector
void v_print(
	const char *		label,
	v_t			a,
	iter_t		m
	
)
{
	iter_t		i;
	
	printf( "%s =\n", label );
	for( i=0 ; i<m ; i++ )
	printf( " %lf", a[i] );

	printf( "\n\r" );
}
